import pytest
from anchor_gpt import PromptLogger, Prompt

def test_log():
    prompt_logger = PromptLogger()
    my_prompt = prompt_logger.log(Prompt(
        text='test prompt',
        response='test response',
        scores={'grounding_distances': [0.1, 0.2, 0.3]},
        embeddings=[0.1, 0.2, 0.3],
    ))

    retreived_prompt = prompt_logger.store.get_by_ids([my_prompt.id])[0]

    assert retreived_prompt.text == 'test prompt'
    assert retreived_prompt.response == 'test response'
    assert retreived_prompt.scores == {'grounding_distances': [0.1, 0.2, 0.3]}
    assert retreived_prompt.embeddings == [0.1, 0.2, 0.3]

    prompt_logger.store.purge()

def test_update():
    prompt_logger = PromptLogger()
    my_prompt = prompt_logger.log(Prompt(
        text='test prompt',
        response='test response',
        scores={'grounding_distances': [0.1, 0.2, 0.3]},
        embeddings=[0.1, 0.2, 0.3],
    ))

    my_prompt.update_scores({'user_feedback': 0.5})

    retreived_prompt = prompt_logger.store.get_by_ids([my_prompt.id])[0]
    assert retreived_prompt.scores == {'grounding_distances': [0.1, 0.2, 0.3], 'user_feedback': 0.5}

    my_prompt.update_scores({'user_feedback': 0.6})
    retreived_prompt = prompt_logger.store.get_by_ids([my_prompt.id])[0]
    assert retreived_prompt.scores == {'grounding_distances': [0.1, 0.2, 0.3], 'user_feedback': 0.6}

    prompt_logger.store.purge()

def test_retrieve():
    prompt_logger = PromptLogger()
    prompt_logger.log(Prompt(
        text='test prompt',
        response='test response',
        scores={'grounding_distances': 0.1},
        embeddings=[0.1, 0.2, 0.3],
    ))

    prompt_logger.log(Prompt(
        text='test prompt 2',
        response='test response',
        scores={'grounding_distances': 0.5},
        embeddings=[0.1, 0.2, 0.3],
    ))

    prompt_logger.log(Prompt(
        text='test prompt 3',
        response='test response',
        scores={'grounding_distances': 0.2},
        embeddings=[0.1, 0.2, 0.3],
    ))


    def retiever(store, threshold):
        def prompt_score(prompt):
            return prompt.scores['grounding_distances']
        return list(filter(lambda x: prompt_score(x) > threshold, store.select_prompts()))

    retreived_prompt = prompt_logger.retrieve(retiever, 0.3)[0]

    assert retreived_prompt.text == 'test prompt 2'

    prompt_logger.store.purge()


def test_dedup():
    prompt_logger = PromptLogger()
    prompt_logger.log(Prompt(
        text='test prompt',
        response='test response',
        scores={'grounding_distances': 0.1},
        embeddings=[0.1, 0.2, 0.3],
    ))

    prompt_logger.log(Prompt(
        text='test prompt 2',
        response='test response',
        scores={'grounding_distances': 0.5},
        embeddings=[-0.3, 0.5, 0.3],
    ))

    prompt_logger.log(Prompt(
        text='test prompt 3',
        response='test response',
        scores={'grounding_distances': 0.2},
        embeddings=[0.9, 0.2, 0.9],
    ))

    prompt_logger.log(Prompt(
        text='test prompt',
        response='test response',
        scores={'grounding_distances': 0.2},
        embeddings=[0.1, 0.2, 0.3],
    ))

    prompt_logger.log(Prompt(
        text='test prompt',
        response='test response',
        scores={'grounding_distances': 0.2},
        embeddings=[0.1, 0.2, 0.3],
    ))

    prompt_logger.log(Prompt(
        text='test prompt',
        response='test response',
        scores={'grounding_distances': 0.2},
        embeddings=[0.1, 0.2, 0.3],
    ))


    def retiever(store, threshold):
        def prompt_score(prompt):
            return prompt.scores['grounding_distances']
        return list(filter(lambda x: prompt_score(x) > threshold, store.select_prompts()))

    retreived_prompts = prompt_logger.retrieve(retiever, 0)

    dedup_prompt = prompt_logger.deduplicate(retreived_prompts, 3)

    dedup_text = [prompt.text for prompt in dedup_prompt]
    dedup_text.sort()
    assert dedup_text == ['test prompt', 'test prompt 2', 'test prompt 3']

    prompt_logger.store.purge()



import pandas as pd
from .prompt_store import prompt_to_db, db_to_prompt, SQLitePromptStore
from .coreset import coreset

class PromptLogger:
    def __init__(self, store=None):
        """
        A prompt logger that stores prompts in a PromptStore.

        Parameters:
            store: a PromptStore object. If None, a SQLitePromptStore will be used.
        """
        self.store = store or SQLitePromptStore()

    def log(self, prompt):
        """
        Log a prompt to the store.

        Parameters:
            prompt: a Prompt object.
        """
        prompt._set_store(self.store)

        return self.store.add(prompt)

    def retrieve(self, retriever, threshold):
        """
        Retrieve the top n prompts according to the retriever function.

        Parameters:
            retriever: a function that takes a PromptStore and threshold a number and returns a list of prompts.
            threshold: a theshold to be consumed by the retreiver.
        """
        return retriever(self.store, threshold)

    def deduplicate(self, prompts, size, distance='euclidean'):
        """
        Remove text duplicates and embeddings near duplicates using the coreset algorithm.
        More details on the coreset algorithm can be found here: https://arxiv.org/abs/1708.00489

        Parameters:
            prompts: a list of Prompt objects.
            size: the number of prompts to select.
            distance: the distance function to use for the coreset algorithm (default: euclidean).
        """

        data = [{'id': prompt.id, 'text': prompt.text} for prompt in prompts]

        df = pd.DataFrame(data)
        df = df.drop_duplicates(subset=['text'])

        prompt_ids = list(df['id'])
        selected_indexes = coreset(
            vector_ids=prompt_ids,
            already_selected=[],
            number_of_datapoints=size,
            transformer=lambda ids: [p.embeddings for p in self.store.get_by_ids(ids)],
            distance=distance
        )

        selected_ids = [prompt_ids[index] for index in selected_indexes]

        return self.store.get_by_ids(selected_ids)



###
# Implementation of the Coreset Algirithm
# Heavily inspired by https://github.com/google/active-learning
###


import os
from datetime import datetime
import numpy as np
from tqdm import tqdm
import random

import tempfile

random.seed(1234)
np.random.seed(1234)

try:
    # Using cuML to GPU enable coreset, otherwise use sklearn
    from cuml.metrics.pairwise_distances import pairwise_distances
except Exception as e:
    from sklearn.metrics import pairwise_distances

BATCH_SIZE = int(os.environ.get('CORESET_BATCH_SIZE', '100'))

class kCenterGreedy():

  def __init__(self, X, cache_dir, vector_shape, batch_shapes, metric='euclidean'):
    self.X = X
    self.metric = metric
    self.min_distances = None
    self.already_selected = []
    self.cache_dir = cache_dir
    self.vector_shape = vector_shape
    self.batch_shapes = batch_shapes

  def get_vectors(self, uuids):
    vector_array = np.empty(shape=(len(uuids), *self.vector_shape), dtype='float16')
    for index, uuid in enumerate(uuids):
      vector_array[index] = np.memmap(
        os.path.join(self.cache_dir, str(uuid)), dtype='float16', mode='r', shape=self.vector_shape)
    return vector_array

  def get_batches(self, batch_start):
    return np.memmap(
        os.path.join(self.cache_dir, str(batch_start)), dtype='float16', mode='r',
        shape=self.batch_shapes[batch_start])


  def update_distances(self, cluster_centers, only_new=True, reset_dist=False):
    if reset_dist:
      self.min_distances = None
    if only_new:
      cluster_centers = [d for d in cluster_centers if d not in self.already_selected]
    cluster_centers = np.array(cluster_centers)
    if len(cluster_centers):
      dist_list = []
      x = self.get_vectors(self.X[cluster_centers])

      for batch_start in range(0, len(self.X), BATCH_SIZE):
        features_batch = self.get_batches(batch_start)
        batch_dist = pairwise_distances(features_batch, x, metric=self.metric)
        dist_list.append(batch_dist)
      dist = np.concatenate(dist_list)

      if self.min_distances is None:
        self.min_distances = np.min(dist, axis=1).reshape(-1,1)
      else:
        self.min_distances = np.minimum(self.min_distances, dist)

  def select_batch_(self, already_selected, N):

    self.update_distances(already_selected, only_new=False, reset_dist=True)

    new_batch = []

    n_obs = len(self.X)

    for _ in tqdm(range(N), desc='calculating coreset...'):
      if self.min_distances is None:
        # Initialize centers with a randomly selected datapoint
        ind = np.random.choice(np.arange(n_obs))
      else:
        ind = np.argmax(self.min_distances)
      # New examples should not be in already selected since those points
      # should have min_distance of zero to a cluster center.
      assert ind not in already_selected

      self.update_distances([ind], only_new=True, reset_dist=False)
      new_batch.append(ind)

    self.already_selected = already_selected

    return new_batch

def coreset(vector_ids, already_selected, number_of_datapoints, transformer, distance='euclidean'):

  with tempfile.TemporaryDirectory() as tmpdirname:
    batch_shapes = {}
    for batch_start in range(0, len(vector_ids), BATCH_SIZE):
      features_batch = transformer(vector_ids[batch_start:batch_start + BATCH_SIZE])
      try:
        first_none_vector_idx = features_batch.index(None)
        raise ValueError(f'Cannot use None embeddings for id {vector_ids[first_none_vector_idx + batch_start]} to calculate coreset')
      except ValueError:
        pass

      features_batch = [np.array(f, dtype='float16') for f in features_batch]
      for index, uuid in enumerate(vector_ids[batch_start:batch_start + BATCH_SIZE]):
        vector_shape = features_batch[index].shape
        fp = np.memmap(os.path.join(tmpdirname, str(uuid)), dtype='float16', mode='w+', shape=features_batch[index].shape)
        fp[:] = features_batch[index][:]
        fp.flush()
      try:
        np_features_batch = np.stack(features_batch, axis=0)
      except ValueError as e:
        raise ValueError(f'Cannot use provided embeddings to calculate coreset: {e}') from e

      bfp = np.memmap(
        os.path.join(tmpdirname, str(batch_start)), dtype='float16', mode='w+', shape=np_features_batch.shape)
      bfp[:] = np_features_batch[:]
      bfp.flush()
      batch_shapes[batch_start] = np_features_batch.shape

    X = np.array(vector_ids)

    return kCenterGreedy(
        X=X,
        metric=distance,
        cache_dir=tmpdirname,
        vector_shape=vector_shape,
        batch_shapes=batch_shapes
    ).select_batch_(
        already_selected=already_selected,
        N=min(number_of_datapoints, X.size)
    )


import uuid
import numpy as np

class Prompt:
    def __init__(self, text, response, scores={}, id=None, embeddings=None):
        self.text = text
        self.response = response
        self.id = id or uuid.uuid4()
        self.embeddings = embeddings
        self.scores = scores

    def _set_store(self, store):
        self.store = store

    def update_scores(self, score):
        """
        Add a score to the prompt.

        Parameters:
            score: a dict. Will be passed in the prompts to the PromptLogger retriever function to allow custom scoring.
        """
        self.scores.update(score)
        self.store.update(self)

    def set_embeddings(self, embeddings):
        """
        Set the embeddings of the prompt.

        Parameters:
            embeddings: a list of floats.
        """
        self.embeddings = embeddings
        self.store.update(self)

    def __repr__(self):
        return f'Prompt(id={self.id}, text={self.text}, response={self.response}, shape(embeddings)={np.shape(self.embeddings) if self.embeddings else None}, scores={self.scores})'



import sqlite3
import uuid
import ast

sqlite3.register_adapter(uuid.UUID, lambda u: u.bytes_le)

from .prompt import Prompt

class PromptStore:
    """
    An extensible class for prompt stores.
    """
    def add(self, prompt):
        """
        Add a prompt to the store.

        Parameters:
            prompt: a Prompt object.
        """
        raise NotImplementedError

    def update(self, prompt):
        """
        Update a prompt in the store.

        Parameters:
            prompt: a Prompt object.
        """
        raise NotImplementedError

    def get_by_ids(self, ids):
        """
        Get a list of prompts by their ids.

        Parameters:
            ids: a list of UUIDs.
        """
        raise NotImplementedError

    def purge(self):
        """
        Purge the store of all prompts.

        """
        raise NotImplementedError


def prompt_to_db(prompt):
    """
    Convert a prompt to a tuple that can be inserted into the database.

    Parameters:
        prompt: a Prompt object.
    """
    return (prompt.id, prompt.text, prompt.response, str(prompt.scores) if prompt.scores else None, str(prompt.embeddings) if prompt.embeddings else None)

def db_to_prompt(db_prompt):
    """
    Convert a database row to a Prompt object.

    Parameters:
        db_prompt: a database row.
    """
    return Prompt(
        id=uuid.UUID(bytes_le=db_prompt['id']),
        text=db_prompt['text'],
        response=db_prompt['response'],
        scores=ast.literal_eval(db_prompt['scores']),
        embeddings=ast.literal_eval(db_prompt['embeddings']) if db_prompt['embeddings'] else None
    )

class SQLitePromptStore(PromptStore):
    """
    An impementation of a prompt store that uses a SQLite backend.
    """
    def __init__(self, name='anchor.db', debug=False):
        self.name = name
        self.connection = sqlite3.connect(name)
        if debug:
            self.connection.set_trace_callback(print)
        self.connection.row_factory = sqlite3.Row
        self.connection.execute('CREATE TABLE IF NOT EXISTS prompts (id UUID, text TEXT, response TEXT, scores JSON, embeddings JSON)')

    def add(self, prompt):
        self.connection.execute('INSERT INTO prompts VALUES (?, ?, ?, ?, ?)', prompt_to_db(prompt))
        self.connection.commit()
        return prompt

    def update(self, prompt):
        (id, text, response, scores, embeddings) = prompt_to_db(prompt)
        self.connection.execute('UPDATE prompts SET text=?, response=?, scores=?, embeddings=? WHERE id=?', (text, response, scores, embeddings, id))
        self.connection.commit()
        return prompt

    def purge(self):
        self.connection.execute('DELETE FROM prompts')
        self.connection.commit()

    def get_by_ids(self, ids):
        result = self.connection.execute(f'SELECT * FROM prompts WHERE id IN ({",".join(["?"] * len(ids))})', ids).fetchall()

        return [db_to_prompt(row) for row in result]

    def select_prompts(self, fields='text, response, id, scores, embeddings', where='1=1', order_by='id', limit='-1'):

        return [db_to_prompt(row) for row in self.connection.execute(f'SELECT {fields} FROM prompts WHERE {where} ORDER BY {order_by} LIMIT {limit}')]
